import logging
import os.path
import time
from enum import Enum
from typing import Callable, List, Optional

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.util.orderedset import UniqueFIFO, FIFO

logger = logging.getLogger(os.path.basename(__file__))


class Action(Enum):
    # stop the search
    STOP = 0
    # do not apply visit() and do not accu neighbors
    SKIP = 1
    # normal behavior
    NONE = 3


def _pre_visit_passthrough(
    api: MetaThesaurus, cui: str, distance: int, distances: FIFO
) -> Action:
    return Action.NONE


def _post_visit_passthrough(
    api: MetaThesaurus, cui: str, distance: int, distances: FIFO
) -> Action:
    return Action.NONE


def breadth_first_search(
    api: MetaThesaurus,
    start_cui: str,
    visit: Callable[[MetaThesaurus, str, int], None],
    get_neighbors: Callable[[MetaThesaurus, str], List[str]],
    pre_visit: Optional[Callable[[MetaThesaurus, str, int, FIFO], Action]] = None,
    post_visit: Optional[Callable[[MetaThesaurus, str, int, FIFO], Action]] = None,
) -> None:
    """
    Do a breadth-first search over UMLS, hunting for definitions.

    Neighbors are determined by `get_neighbors`, which defaults to :py:func: umlsrat.lookup.umls.get_broader_concepts

    Resulting umlsrat.lookup.graph_fn.Action of `pre_visit` and `post_visit` allow additional
    control of the search. By default, these do nothing.


    :param api: MetaThesaurus API
    :param start_cui: starting Concept ID
    :param visit: applied to each visited concept
    :param get_neighbors: get neighbors for a given CUI.
    :param pre_visit: applied before visiting a concept. default: do nothing
    :param post_visit: applied after visiting a concept. default: do nothing
    """
    assert api
    assert start_cui

    if pre_visit is None:
        pre_visit = _pre_visit_passthrough

    if post_visit is None:
        post_visit = _post_visit_passthrough

    to_visit = UniqueFIFO([start_cui])
    distances = FIFO([0])

    visited = UniqueFIFO()

    while to_visit:
        current_cui = to_visit.peek()
        current_dist = distances.peek()

        logger.info(
            f"{len(visited)} '{current_cui}' "
            f"curDistance = {current_dist} "
            f"numToVisit = {len(to_visit)} "
        )

        pre_visit_action = pre_visit(api, current_cui, current_dist, distances)
        if pre_visit_action == Action.STOP:
            break

        if pre_visit_action != Action.SKIP:
            visit(api, current_cui, current_dist)

        ## Mark as visited
        visited.push(to_visit.pop())
        distances.pop()

        if pre_visit_action == Action.SKIP:
            continue

        post_visit_action = post_visit(api, current_cui, current_dist, distances)

        if post_visit_action == Action.STOP:
            break

        if post_visit_action == Action.SKIP:
            continue

        ## Find neighbors and add to_visit
        t0 = time.time()
        neighbors = get_neighbors(api, current_cui)
        logger.debug(
            "{} retrieved {} neighbors in {:.3f} sec".format(
                current_cui, len(neighbors), time.time() - t0
            )
        )
        for neighbor_cui in neighbors:
            # remove those we have visited or plan to visit already
            if neighbor_cui in visited or neighbor_cui in to_visit:
                continue
            to_visit.push(neighbor_cui)
            distances.push(current_dist + 1)

    return



from typing import Callable, Tuple, Dict, Set, Any

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.util import text


def _get_sem_types(concept: Dict) -> Set[str]:
    return {_.get("name") for _ in concept.get("semanticTypes")}


def sem_type_name_sim(
    api: MetaThesaurus, source_cui: str
) -> Callable[[str], Tuple[bool, float]]:
    source = api.get_concept(source_cui)
    source_sem_types = _get_sem_types(source)
    hammingish = text.hammingish_partial(source.get("name"))

    def sort_fn(cui: str) -> Tuple[bool, float]:
        target = api.get_concept(cui)
        target_sem_types = _get_sem_types(target)
        overlap = bool(source_sem_types & target_sem_types)
        return not overlap, hammingish(target.get("name"))

    return sort_fn


def cui_name_sim(api: MetaThesaurus, source_cui: str) -> Callable[[str], Any]:
    source_name = api.get_concept(source_cui).get("name")
    if source_name is None:
        mth_atom = list(api.get_atoms_for_cui(source_cui, sabs="MTH")).pop()
        source_name = mth_atom.get("name")
    assert source_name, f"No name for CUI {source_cui}"
    return desc_name_sim(api, source_name)


def desc_name_sim(api: MetaThesaurus, desc: str) -> Callable[[str], Any]:
    hammingish = text.hammingish_partial(desc)

    def sort_fn(cui: str):
        concept = api.get_concept(cui)
        target = concept.get("name")
        return hammingish(target), target

    return sort_fn


def relation_count(api: MetaThesaurus, dsc: bool = False) -> Callable[[str], int]:
    def sort_fn(cui: str) -> int:
        cnt = api.get_concept(cui).get("relationCount")
        return -cnt if dsc else cnt

    return sort_fn



import argparse
import json
import logging
import os
from os.path import expanduser
from typing import Optional, Any, Dict, List, Iterator

from backoff import on_exception, expo
from ratelimit import limits, RateLimitException
from requests import Session, HTTPError, Response
from requests.adapters import HTTPAdapter
from requests_cache import CachedSession
from urllib3 import Retry

from umlsrat import const
from umlsrat.util import args_util


def _set_retries(session):
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[500, 502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session


def _configure_session(session):
    _set_retries(session)
    return session


def _cache_path(cache_name: str) -> str:
    return expanduser(f"~/.cache/umls-rat/{cache_name}")


_NONE = "NONE"


def _interpret_none(value: Any):
    if isinstance(value, str):
        if value == _NONE:
            return None
        else:
            return value
    elif isinstance(value, Dict):
        return {key: _interpret_none(value) for key, value in value.items()}
    elif isinstance(value, List):
        return [_interpret_none(_) for _ in value]
    else:
        return value


def _extract_results(response_json: Dict) -> List[Dict]:
    """
    Extract results from response json.
    :param response_json:
    :return: list of results or None if invalid
    """
    if not response_json:
        return []

    result = response_json["result"]

    if "results" in result:
        return result["results"]
    else:
        return result


class MetaThesaurusSession(object):
    def __init__(self, api_key: Optional[str] = None, use_cache: bool = True):
        """
        Constructor.

        If ``api_key`` is not passed, the value will be read from the ``UMLS_API_KEY`` environment variable.

        :param api_key: API key acquired from `here <https://uts.nlm.nih.gov/uts/signup-login>`__. If not passed as an argument, the value will be read from the ``UMLS_API_KEY`` environment variable.
        :param use_cache: use cache for requests (default ``True``)
        """
        if api_key:
            self._api_key = api_key
        else:
            self._api_key = os.environ.get(const.API_KEY_ENV_VAR)
            if not self._api_key:
                raise KeyError(
                    f"`api_key` not passed and `{const.API_KEY_ENV_VAR}` not set."
                )

        self.use_cache = use_cache

        if self.use_cache:
            self.session = CachedSession(
                cache_name=_cache_path("api-cache"),
                ignored_parameters=["apiKey"],
                allowable_codes=[200, 400, 404],
            )
        else:
            self.session = Session()

        self.session = _configure_session(self.session)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.session.close()

    def get_current_version(self) -> str:
        """
        Get the precise current version of UMLS. (e.g. 2023AA)
        """
        # use a temp session to ensure cache is disabled
        with MetaThesaurusSession(api_key=self._api_key, use_cache=False) as s:
            # get info for "Controlled Vocabulary"
            content_base = const.UMLS_REST_ENDPOINT + "/content/"
            result = s.get_single_result(content_base + "current/CUI/C0282503")
            # parse version from atoms url
            version = result["atoms"][len(content_base) :].split("/")[0]

        return version

    @staticmethod
    def add_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Add constructor arguments to parser.

        :param parser: parser
        :return: the same parser
        """
        group = parser.add_argument_group("Session")
        group.add_argument("--api-key", type=str, help="API key", default=None)
        group.add_argument(
            "--use-cache",
            help="Controls whether or not we use the response cache",
            type=args_util.str2bool,
            default=True,
        )

        return parser

    @staticmethod
    def from_namespace(args: argparse.Namespace) -> "MetaThesaurusSession":
        """
        Construct new object using values from namespace.

        :param args: parsed args
        :return: new session
        """
        return MetaThesaurusSession(api_key=args.api_key, use_cache=args.use_cache)

    @property
    def _logger(self):
        return logging.getLogger(self.__class__.__name__)

    def _get_cached(self, url: str, params) -> Optional[Response]:
        """Get the cached response or None"""
        if not self.use_cache:
            return None

        response = self.session.get(url, params=params, only_if_cached=True)
        if response.status_code == 504:
            return None
        else:
            return response

    @on_exception(expo, RateLimitException, max_tries=8)
    @limits(calls=20, period=1)
    def _get_http(self, url, params):
        """
        Rate limited per `Terms of Service <https://documentation.uts.nlm.nih.gov/terms-of-service.html>`__
        """
        response = self.session.get(url=url, params=params)
        return response

    def get_json(
        self, url: str, strict: Optional[bool] = False, **params
    ) -> Optional[Dict]:
        """
        Get JSON response and interpret None values.

        :param url: target URL
        :param strict: if True raise for status even when expected
        :param params: params to the get call. params with None value are dropped.
        :return: the JSON response
        """
        assert url, "Must provide URL"

        if "apiKey" in params:
            self._logger.warning(
                f"'apiKey' should not be in params! Will be overwritten..."
            )

        # copy and drop None values
        params = {k: v for k, v in params.items() if v is not None}
        # add apiKey
        params["apiKey"] = self._api_key
        try:
            response = self._get_cached(url, params=params)
            if response is None:
                response = self._get_http(url, params=params)
        except Exception as e:
            self._logger.exception("Failed to get %s", url)
            raise e

        try:
            response.raise_for_status()
        except HTTPError as e:
            if strict:
                raise e

            self._logger.debug("Caught HTTPError: %s", e)
            if e.response.status_code == 400 or e.response.status_code == 404:
                # we interpret this as "you're looking for something that isn't there"
                return None
            else:
                raise e

        return _interpret_none(response.json())

    def get_results(
        self,
        url: str,
        max_results: Optional[int] = None,
        **params,
    ) -> Iterator[Dict]:
        """
        Get data from arbitrary URI. Will return an empty list on 400 or 404
        unless `strict=True` is in kwargs, in which case it will raise.

        :param url: URL under http://uts-ws.nlm.nih.gov/rest
        :param max_results: maximum number of result to return. None = no max
        :param params: parameters sent with the get request
        :return: generator yielding results
        """
        assert url
        if max_results is not None:
            assert max_results > 0, "max_results must be > 0"

        response_json = self.get_json(url, **params)
        if not response_json:
            return

        # get page count from first call
        page_count = response_json.get("pageCount", None)

        if "pageNumber" not in response_json:
            raise ValueError(
                "Expected pagination fields in response:\n"
                "{}".format(json.dumps(response_json, indent=2))
            )

        if "pageNumber" not in params:
            params["pageNumber"] = 1
        else:
            self._logger.warning(
                "Starting pagination from page %d", params["pageNumber"]
            )

        n_yielded = 0
        while True:
            results = _extract_results(response_json)
            if not results:
                return

            for r in results:
                yield r
                n_yielded += 1
                if n_yielded == max_results:
                    return

            if page_count == params["pageNumber"]:
                # no more pages
                return

            # next page
            params = dict(pageNumber=params.pop("pageNumber") + 1, **params)
            response_json = self.get_json(url, **params)

    def get_single_result(self, url: str, **params) -> Optional[Dict]:
        """
        When you know there will only be one coming back

        :param url: URL under http://uts-ws.nlm.nih.gov/rest
        :param params: parameters sent with the get request
        :return: a result or None
        """
        response_json = self.get_json(url, **params)
        if not response_json:
            return

        return response_json["result"]



import textwrap
from itertools import cycle, islice
from typing import List, Dict, Iterable

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.lookup import lookup_umls


def map_cuis_to_names(api: MetaThesaurus, cuis: Iterable[str]) -> List[str]:
    """
    Map CUIs to concept names.
    :param api: meta thesaurus object
    :param cuis: CUIs
    :return: names
    """
    return [lookup_umls.get_concept_name(api, cui) for cui in cuis]


def extract_concept_names(concepts: List[Dict]) -> List[str]:
    """
    Extract names from concepts.
    :param concepts: concept dicts
    :return: names
    """
    return [_["name"] for _ in concepts]


def extract_definitions(concepts: List[Dict]) -> List[str]:
    """
    Extract definition text from concepts.
    :param concepts: concept dicts
    :return: definitions
    """
    defs = []
    for concept in concepts:
        for d in concept["definitions"]:
            defs.append(d)
    return [d["value"] for d in defs]


def _entry_to_string(name: str, definitions: List[Dict]) -> str:
    value = ""
    value += f"{name}\n"
    value += "=" * len(name)
    value += "\n"
    enum_defs = (
        textwrap.fill(f"{x + 1}. {datum['value']}")
        for x, datum in enumerate(definitions)
    )
    value += "\n".join(enum_defs)
    return value


def definitions_to_md(concepts: List[Dict]) -> str:
    """
    Write list of defined concepts as MarkDown.

    :param concepts: list of concept Dicts
    :return: MarkDown string
    """
    entries = (_entry_to_string(c["name"], c["definitions"]) for c in concepts)
    return "\n\n".join(entries)


def roundrobin(*iterables):
    "roundrobin('ABC', 'D', 'EF') --> A D E B F C"
    # Recipe credited to George Sakkis
    num_active = len(iterables)
    nexts = cycle(iter(it).__next__ for it in iterables)
    while num_active:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            # Remove the iterator we just exhausted from the cycle.
            num_active -= 1
            nexts = cycle(islice(nexts, num_active))



import json
import logging
import os.path
from typing import Optional, Dict, List, Iterator, Iterable

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.util import text, cui_order

logger = logging.getLogger(os.path.basename(__file__))


def get_concept_name(api: MetaThesaurus, cui: str) -> Optional[str]:
    """
    Get the name for a CUI.

    :param api: api object
    :param cui: the cui
    :return: concept name or None if it's not set
    """
    return api.get_concept(cui).get("name", None)


def _term_search(api: MetaThesaurus, term: str, max_results: int) -> Dict:
    for search_type in ("words", "normalizedWords", "approximate"):
        for input_type in ("sourceConcept", "sourceDescriptor"):
            search_params = dict(
                inputType=input_type,
                searchType=search_type,
                pageSize=min(25, max_results),
            )
            concepts = api.search(term, max_results=max_results, **search_params)
            # filter bogus results
            concepts = [_ for _ in concepts if _["ui"]]
            if concepts:
                # sort by hammingish -- the UMLS search is probably better in most cases
                # txt_key = text.hammingish_partial(term)
                #
                # def sort_key(c: Dict) -> float:
                #     return txt_key(c["name"])
                #
                # concepts = sorted(concepts, key=sort_key)
                return dict(**search_params, searchTerm=term, concepts=concepts)

    return dict(searchTerm=term, concepts=[])


def _is_strict_match(original: str, matched: str) -> bool:
    o_tokens = text.norm_tokenize(original)
    m_tokens = text.norm_tokenize(matched)

    # if each is only one token and one contains the other exactly
    if len(o_tokens) == len(m_tokens) == 1 and (
        o_tokens[0] in m_tokens[0] or m_tokens[0] in o_tokens[0]
    ):
        return True

    hdist = text.hammingish(o_tokens, m_tokens)
    return hdist < 0.5


def term_search(
    api: MetaThesaurus,
    term: str,
    max_results: Optional[int] = 1000,
    strict_match: Optional[bool] = False,
) -> Dict:
    """
    Search for a term in UMLS. Increase the flexibility of the search with each iteration.

    Pseudo-Python code looks like this:

    .. code-block:: python

        for search_type in ("words", "normalizedWords", "approximate"):
            for input_type in ( "sourceConcept", "sourceDescriptor"):
                # Search for the term with above params.
                api.search(term, ...)
                # If found, return.
                ...

    See: :py:func:`umlsrat.api.metathesaurus.MetaThesaurus.search`

    :param api: api object
    :param term: the term to search for
    :param max_results: maximum number of results (default: 1000)
    :param strict_match: only allow strict matches (default: False)
    :return: search result
    """
    result = _term_search(api, term, max_results)

    if strict_match:
        result["concepts"] = [
            concept
            for concept in result["concepts"]
            if _is_strict_match(result["searchTerm"], concept["name"])
        ]

    return result


def _do_cui_search(
    api: MetaThesaurus, source_vocab: str, concept_id: str, **kwargs
) -> List[str]:
    search_params = dict(
        inputType="sourceUi",
        searchType="exact",
        sabs=source_vocab,
    )
    search_params.update(kwargs)
    results = list(api.search(query=concept_id, **search_params))

    cuis = [_["ui"] for _ in results]

    ordered = sorted(cuis, key=cui_order.relation_count(api, dsc=False))
    return ordered


def get_cuis_for(api: MetaThesaurus, source_vocab: str, concept_id: str) -> List[str]:
    """
    Get UMLS CUIs for a source concept.

    **Remember** there can be more than one CUI associated with
    a given source concept.

    Try the following in order until we find some CUIs:

    #. Search for this concept using search API
    #. Same search including Obsolete and Suppressible

    :param api: MetaThesaurus
    :param source_vocab: source vocabulary e.g. SNOMED
    :param concept_id: concept ID in the source vocab
    :return: CUI or None if not found
    """

    assert concept_id
    source_vocab = api.validate_source_abbrev(source_vocab)

    cuis = _do_cui_search(api, source_vocab, concept_id)
    if cuis:
        return cuis

    ## might have an obsolete concept
    cuis = _do_cui_search(
        api, source_vocab, concept_id, includeObsolete=True, includeSuppressible=True
    )
    if cuis:
        return cuis

    # we cannot find any CUIs for this concept
    return []


def _concepts_for_obj(api: MetaThesaurus, obj: Dict) -> List[Dict]:
    if obj.get("classType") == "Concept":
        return [obj]

    if "concept" in obj:
        return [api.session.get_single_result(obj["concept"])]
    elif "concepts" in obj:
        return api.session.get_results(obj["concepts"])
    else:
        raise ValueError(f"No concepts for atom:\n{obj}")


def _get_source_related_concepts(
    api: MetaThesaurus,
    cui: str,
    allowed_relation_str: str,
    language: str = None,
    **add_params,
) -> Iterator[Dict]:
    # get all atoms associated with this umls concept
    base_atoms = api.get_atoms_for_cui(cui, language=language, **add_params)

    for base_atom in base_atoms:
        if base_atom.get("rootSource") == "MTH":
            # place-holder atoms for metathesaurus
            continue

        code_url = base_atom["code"]
        if not code_url:
            raise ValueError("'code' is not available")

        if code_url.endswith("NOCODE"):
            # this is strange
            logger.debug(
                "NOCODE associated with atom:\n%s", json.dumps(base_atom, indent=2)
            )
            continue

        source_code_atom = api.session.get_single_result(code_url)
        if not source_code_atom:
            logger.warning(
                f"Got null code atom for {code_url} from\n{json.dumps(base_atom, indent=2)}"
            )
            # dead end
            continue

        source_relations = list(
            api.get_source_relations(
                source_vocab=source_code_atom["rootSource"],
                concept_id=source_code_atom["ui"],
                includeRelationLabels=allowed_relation_str,
                **add_params,
            )
        )

        for rel in source_relations:
            if "relatedId" not in rel:
                # dead end
                continue

            related_atom = api.session.get_single_result(rel["relatedId"])

            for related_concept in _concepts_for_obj(api, related_atom):
                yield related_concept


def _get_related_cuis(
    api: MetaThesaurus,
    cui: str,
    allowed_relations: Iterable[str],
    include_source_relations: bool,
    sabs: str,
    language: str,
    **add_params,
) -> Iterator[str]:
    """
    Get related concepts -- either direct (UMLS) relations or those related in a source vocab.
    **NO GUARANTEES REGARDING RETURN ORDER**

    :param api: meta thesaurus
    :param cui: starting concept CUI
    :param allowed_relations: relations of interest
    :param language: target language
    :param add_params: additional parameters passed to internal calls
    :return: generator yielding CUIs
    """

    seen = {cui}

    def yield_unique_concept_ui(next_concept: Dict):
        if not next_concept:
            return

        next_cui = next_concept["ui"]
        if not next_cui:
            return

        if next_cui not in seen:
            yield next_cui
            seen.add(next_cui)

    sabs_str = api.get_sabs_str(language, sabs)

    allowed_relation_str = ",".join(sorted(allowed_relations))

    ##
    # First get direct relations
    ##
    for rel in api.get_relations(
        cui,
        sabs=sabs_str,
        includeRelationLabels=allowed_relation_str,
        **add_params,
    ):
        atom = api.session.get_single_result(rel["relatedId"])
        for concept in _concepts_for_obj(api, atom):
            yield from yield_unique_concept_ui(concept)

    if not include_source_relations:
        return

    ##
    # Find all concepts related via an associated source atom;
    ##
    for related_concept in _get_source_related_concepts(
        api,
        cui=cui,
        allowed_relation_str=allowed_relation_str,
        language=language,
        **add_params,
    ):
        yield from yield_unique_concept_ui(related_concept)


def get_related_cuis(
    api: MetaThesaurus,
    cui: str,
    allowed_relations: Iterable[str],
    include_source_relations: bool = False,
    language: str = None,
    sabs: str = None,
) -> List[str]:
    """
    Get CUIs for related concepts.

    Concepts must be related via ``allowed_relations``. If no related concepts are
    found, try Obsolete and Suppressible.

    CUIs are returned in a fixed order.


    :param api: meta thesaurus
    :param cui: starting concept CUI
    :param allowed_relations: relations of interest
    :param include_source_relations: include CUIs related by atomic source relations
    :param language: target language
    :param sabs: comma separated list of source abbreviations

    :return: list of CUIs
    """

    related_cuis = list(
        _get_related_cuis(
            api=api,
            cui=cui,
            allowed_relations=allowed_relations,
            include_source_relations=include_source_relations,
            sabs=sabs,
            language=language,
        )
    )

    if not related_cuis:
        related_cuis = list(
            _get_related_cuis(
                api=api,
                cui=cui,
                allowed_relations=allowed_relations,
                include_source_relations=include_source_relations,
                sabs=sabs,
                language=language,
                includeObsolete=True,
                includeSuppressible=True,
            )
        )

    ordered = sorted(related_cuis)
    return ordered


def get_broader_cuis(
    api: MetaThesaurus,
    cui: str,
    include_source_relations: bool = False,
    language: str = None,
    sabs: str = None,
) -> List[str]:
    """
    Get CUIs for broader, related concepts.

    CUIs are returned in a fixed order.


    :param api: meta thesaurus
    :param cui: starting concept CUI
    :param include_source_relations: include CUIs related by atomic source relations
    :param language: target language
    :param sabs: comma separated list of source abbreviations
    :return: list of CUIs
    """
    return get_related_cuis(
        api=api,
        cui=cui,
        allowed_relations=("RN", "CHD"),
        include_source_relations=include_source_relations,
        language=language,
        sabs=sabs,
    )


def get_narrower_cuis(
    api: MetaThesaurus,
    cui: str,
    include_source_relations: bool = False,
    language: str = None,
    sabs: str = None,
) -> List[str]:
    """
    Get CUIs for narrower, related concepts.

    CUIs are returned in a fixed order.

    :param api: meta thesaurus
    :param cui: starting concept CUI
    :param include_source_relations: include CUIs related by atomic source relations
    :param language: target language
    :param sabs: comma separated list of source abbreviations
    :return: list of CUIs
    """
    return get_related_cuis(
        api=api,
        cui=cui,
        allowed_relations=("RB", "PAR"),
        include_source_relations=include_source_relations,
        language=language,
        sabs=sabs,
    )



from typing import List, Callable

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.lookup import lookup_umls
from umlsrat.util import text
from umlsrat.util.orderedset import UniqueFIFO


def _get_norm_fn(normalize: bool) -> Callable[[str], str]:
    if normalize:
        return text.normalize
    else:
        return text.identity


def _append_cui_descriptions(
    api: MetaThesaurus,
    cui: str,
    syn_names: UniqueFIFO,
    txt_norm: Callable[[str], str],
    sabs: str,
) -> List[str]:
    def push_name(name: str) -> None:
        syn_names.push(txt_norm(name))

    for atom in api.get_atoms_for_cui(cui, sabs=sabs):
        push_name(atom.get("name"))

    return syn_names.items


def get_synonyms(
    api: MetaThesaurus,
    cui: str,
    language: str = "ENG",
    normalize: bool = False,
) -> List[str]:
    """
    Find unique, synonymous concept names. Return order is determined by the API. Uniqueness is determined by
    case-insensitive exact string match or by the normalized form if ``normalize=True``.

    >>> from umlsrat.api.metathesaurus import MetaThesaurus
    >>> from umlsrat.lookup.lookup_desc import get_synonyms
    >>> get_synonyms(MetaThesaurus(), "C0034500")

    .. code-block:: js

        [
            "raccoon",
            "Procyon",
            "Procyon, NOS",
            "raccoons",
            "Procyons",
            "Genus Procyon",
            "Genus Procyon (organism)",
        ]


    :param api: MetaThesaurus
    :param cui: CUI
    :param language: target language
    :param normalize: normalize names
    :return: list of names for this concept
    """
    # get source vocab abbreviations for language
    sabs = api.get_sabs_str(language)
    # create name queue
    syn_names = UniqueFIFO(keyfn=str.lower)
    # get text norm function
    txt_norm = _get_norm_fn(normalize)
    _append_cui_descriptions(
        api=api, cui=cui, syn_names=syn_names, txt_norm=txt_norm, sabs=sabs
    )
    return syn_names.items


def find_synonyms(
    api: MetaThesaurus,
    source_vocab: str,
    concept_id: str,
    language: str = "ENG",
    normalize: bool = False,
) -> List[str]:
    """
    Find unique, synonymous concept names given a source concept. Return order is determined by
    the API. Uniqueness is determined by case-insensitive exact string match or by the normalized
    form if ``normalize=True``.

    >>> from umlsrat.api.metathesaurus import MetaThesaurus
    >>> from umlsrat.lookup.lookup_desc import find_synonyms
    >>> find_synonyms(MetaThesaurus(), "ICD10CM", "T87.44")

    .. code-block:: js

        [
          "Infection of amputation stump, left lower extremity",
          "infection of amputation stump of left lower extremity",
          "infection of amputation stump of left lower extremity (diagnosis)",
          "Infection of amputation stump of left lower limb",
          "Infection of amputation stump of left leg",
          "Infection of amputation stump of left lower limb (disorder)"
        ]

    With normalization.

    >>> find_synonyms(MetaThesaurus(), "ICD10CM", "T87.44", normalize=True)

    .. code-block:: js

        [
          "infection of amputation stump left lower extremity",
          "infection of amputation stump of left lower extremity",
          "infection of amputation stump of left lower limb",
          "infection of amputation stump of left leg"
        ]

    :param api: MetaThesaurus
    :param source_vocab: source vocabulary e.g. ICD10CM
    :param concept_id: concept ID in the source vocab
    :param language: target language
    :param normalize: normalize names
    :return: list of names for this concept
    """
    # get source vocab abbreviations for language
    sabs = api.get_sabs_str(language)
    # create name queue
    syn_names = UniqueFIFO(keyfn=str.lower)
    # get text norm function
    txt_norm = _get_norm_fn(normalize)

    base_concept = api.get_source_concept(source_vocab, concept_id)
    if not base_concept:
        raise ValueError(f"Source concept not found {source_vocab}/{concept_id}")

    if source_vocab in sabs.split(","):
        # The name of the base concept always comes first -- provided that it is a vocab associated
        # with the desired language.
        syn_names.push(txt_norm(base_concept.get("name")))

    for cui in lookup_umls.get_cuis_for(api, source_vocab, concept_id):
        _append_cui_descriptions(
            api=api,
            cui=cui,
            txt_norm=txt_norm,
            syn_names=syn_names,
            sabs=sabs,
        )

    return syn_names.items



UMLS_REST_ENDPOINT = "https://uts-ws.nlm.nih.gov/rest"

API_KEY_ENV_VAR = "UMLS_API_KEY"

# Common vocabulary name mapping
_COMMON_VOCAB_NAMES = {
    "SNOMED": "SNOMEDCT_US",
    "SNOMED-CT": "SNOMEDCT_US",
    "LOINC": "LNC",
}



import argparse
import collections
import itertools
import logging
import operator
import os.path
from typing import Optional, Iterable, List, Dict, Callable, Set, Any

from umlsrat.api.metathesaurus import MetaThesaurus
from umlsrat.lookup import graph_fn, lookup_umls
from umlsrat.lookup.graph_fn import Action
from umlsrat.util import orderedset, text, iterators
from umlsrat.util.args_util import str2bool
from umlsrat.util.orderedset import FIFO

logger = logging.getLogger(os.path.basename(__file__))


def _resolve_semantic_types(api: MetaThesaurus, concept: Dict) -> Set[str]:
    """Resolve semantic type data and return the set of type names assigned to this concept"""
    type_names = set()

    sem_types = concept["semanticTypes"]
    if not sem_types:
        return type_names

    for st in sem_types:
        type_names.add(st["name"])
        st_uri = st.pop("uri", None)
        if st_uri:
            st["data"] = api.session.get_single_result(st_uri)

    return type_names


def _clean_definition(definition: Dict) -> Dict:
    # clean "value"
    return {
        k: text.clean_definition_text(v) if k == "value" else v
        for k, v in definition.items()
    }


def _definitions_bfs(
    api: MetaThesaurus,
    start_cui: str,
    get_neighbors: Callable[[MetaThesaurus, str], List[str]],
    stop_on_found: Optional[bool] = True,
    max_distance: int = None,
    target_vocabs: Optional[Iterable[str]] = None,
    language: str = "ENG",
    preserve_semantic_type: bool = False,
) -> List[Dict]:
    """
    Do a breadth-first search over UMLS for concepts with associated definitions.


    Resulting concepts
    are returned in order of closest to farthest (relative to ``start_cui``). Within each concept,
    definitions are sorted by length (shortest to longest).


    :param get_neighbors: fn to get neighbors of a CUI (e.g. get_broader or get_narrower)
    :param api: MetaThesaurus API
    :param start_cui: starting Concept ID
    :param stop_on_found: stop searching after processing first level containing defined concepts
    :param max_distance: maximum allowed distance from `start_cui` (0/None = Infinity)
    :param target_vocabs: only allow definitions from these vocabularies
    :param language: target language
    :param preserve_semantic_type: preserve the semantic type assigned to ``start_cui``
    :return: a list of Concepts with Definitions
    """
    assert api
    assert start_cui
    if max_distance is None:
        max_distance = 0
    assert max_distance >= 0
    if not stop_on_found and not max_distance:
        logger.warning(
            f"stop_on_found = {stop_on_found} and max_distance = {max_distance}; this "
            f"could result in an unintentionally massive search space. Recommend setting "
            f"`max_distance=2` and go from there."
        )

    # first get target vocabs based on language
    if target_vocabs:
        target_vocabs = set(target_vocabs)
        for tv in target_vocabs:
            info = api.find_source_info(tv)
            assert (
                info.get("language").get("abbreviation") == language
            ), f"Requested vocabulary {tv} is not in the target language {language}"
    else:
        target_vocabs = set(api.sources_for_language(language))
        assert target_vocabs, f"No vocabularies for language abbreviation '{language}'"

    # If we want to preserve semantic type, we need to get them for the start CUI

    if preserve_semantic_type:
        start_sem_types = _resolve_semantic_types(api, api.get_concept(start_cui))

        def allowed_sem_type(concept: Dict) -> bool:
            sem_types = set(d.get("name") for d in concept.get("semanticTypes"))
            return bool(start_sem_types & sem_types)

    else:
        start_sem_types = None

        def allowed_sem_type(concept: Dict) -> bool:
            return True

    ##
    # visit will accumulate definitions
    ##
    defined_concepts = []

    def visit(api: MetaThesaurus, current_cui: str, current_dist: int):
        current_concept = api.get_concept(current_cui)

        # if this url is not populated, there certainly are no definitions
        if not current_concept["definitions"]:
            return

        definitions = list(api.get_definitions(current_cui))

        # filter defs not in target vocab
        if target_vocabs:
            definitions = [_ for _ in definitions if _["rootSource"] in target_vocabs]

        if not definitions:
            return

        # clean text, sort by length, and drop duplicates
        cleaned = orderedset.UniqueFIFO(
            sorted(map(_clean_definition, definitions), key=lambda _: len(_["value"])),
            keyfn=lambda _: f"{_['rootSource']}/{text.normalize(_['value'])}",
        )

        current_concept["definitions"] = cleaned.items
        _resolve_semantic_types(api, current_concept)
        current_concept["distanceFromOrigin"] = current_dist

        # only add concepts with an allowed semantic type
        if not allowed_sem_type(current_concept):
            return

        # reorder dict for readability
        reordered_concept = collections.OrderedDict(
            (k, current_concept.pop(k))
            for k in (
                "classType",
                "ui",
                "name",
                "distanceFromOrigin",
                "definitions",
            )
        )
        reordered_concept.update(current_concept)

        # add to defined concepts
        defined_concepts.append(reordered_concept)

    ##
    # post visit actions are based on number of definitions and current distance from origin
    def post_visit(
        api: MetaThesaurus, current_cui: str, current_dist: int, distances: FIFO
    ) -> Action:
        if stop_on_found and defined_concepts:
            if not distances or distances.peek() > current_dist:
                return Action.STOP
            else:
                return Action.SKIP

        if max_distance and current_dist == max_distance:
            return Action.SKIP

        return Action.NONE

    # here we actually do the search
    graph_fn.breadth_first_search(
        api=api,
        start_cui=start_cui,
        visit=visit,
        get_neighbors=get_neighbors,
        post_visit=post_visit,
    )

    return defined_concepts


def definitions_bfs(
    api: MetaThesaurus,
    start_cui: str,
    broader: Optional[bool] = True,
    stop_on_found: Optional[bool] = True,
    max_distance: Optional[int] = None,
    target_vocabs: Optional[Iterable[str]] = None,
    language: Optional[str] = "ENG",
    preserve_semantic_type: Optional[bool] = False,
) -> List[Dict]:
    """
    Do a breadth-first search over UMLS for concepts with associated definitions.

    Resulting concepts are returned in order of closest to farthest (relative to ``start_cui``).
    Within each concept, definitions are sorted by length (shortest to longest).

    Use this method when you have a CUI. If you do not have a CUI,
    use :meth:`umlsrat.lookup.definitions.find_defined_concepts`

    :param api: MetaThesaurus API
    :param start_cui: starting Concept ID
    :param broader: search broader concepts. If false, search narrower.
    :param stop_on_found: stop searching after processing first level containing defined concepts
    :param max_distance: maximum allowed distance from `start_cui` (Default = None = 0 = Infinity)
    :param target_vocabs: only allow definitions from these vocabularies
    :param language: only consider concepts in this language
    :param preserve_semantic_type: preserve the semantic type assigned to ``start_cui``
    :return: a list of Concepts with Definitions
    """

    def get_neighbors(api: MetaThesaurus, cui: str) -> List[str]:
        if broader:
            return lookup_umls.get_broader_cuis(api, cui, language=language)
        else:
            return lookup_umls.get_narrower_cuis(api, cui, language=language)

    return _definitions_bfs(
        api=api,
        start_cui=start_cui,
        get_neighbors=get_neighbors,
        stop_on_found=stop_on_found,
        max_distance=max_distance,
        target_vocabs=target_vocabs,
        language=language,
        preserve_semantic_type=preserve_semantic_type,
    )


def find_defined_concepts(
    api: MetaThesaurus,
    source_vocab: str = None,
    concept_id: str = None,
    source_desc: str = None,
    broader: Optional[bool] = True,
    stop_on_found: Optional[bool] = True,
    max_distance: Optional[int] = None,
    language: str = "ENG",
    preserve_semantic_type: bool = False,
) -> List[Dict]:
    """
    Find defined concepts in UMLS which are equivalent to or *broader* than the source concept.

    Resulting concepts are returned in order of closest to farthest (relative to ``start_cui``).
    Within each concept, definitions are sorted by length (shortest to longest).

    Use this function when you do not have a UMLS CUI. If you do have a CUI, you can use
    :meth:`umlsrat.lookup.definitions.broader_definitions_bfs`.

    1. Look up CUIs
        * use the source UI (if given). if not found,
        * use the source description
    2. Run :meth:`umlsrat.lookup.definitions.broader_definitions_bfs` for resulting CUIs

    :param api: MetaThesaurus API
    :param source_vocab: source vocab
    :param concept_id: source concept id
    :param source_desc: source concept description
    :param broader: search broader concepts. If false, search narrower.
    :param stop_on_found: stop searching after processing first level containing defined concepts
    :param max_distance: stop searching after reaching this distance from the original source \
    concept. (Default = None = Infinity)
    :param language: only consider concepts in this language
    :param preserve_semantic_type: only search concept which have the same semantic type as the starting concept
    :return: a list of Concepts with Definitions
    """

    if concept_id:
        assert source_vocab, f"Must provide source vocab for code {concept_id}"
    else:
        assert (
            source_desc
        ), "Must provide either source code and vocab or descriptor (source_desc)"

    if logger.isEnabledFor(logging.INFO):
        msg = f"Finding {language} definition(s) of"
        if concept_id:
            msg = f"{msg} {source_vocab}/{concept_id}"

        if source_desc:
            msg = f"{msg} [{source_desc}]"

        logger.info(msg)

    def definitions_search(start_cui: str) -> List[Dict]:
        data = definitions_bfs(
            api,
            start_cui=start_cui,
            broader=broader,
            stop_on_found=stop_on_found,
            max_distance=max_distance,
            language=language,
            preserve_semantic_type=preserve_semantic_type,
        )

        return data

    if concept_id:
        cuis_from_code = lookup_umls.get_cuis_for(
            api, source_vocab=source_vocab, concept_id=concept_id
        )
        if cuis_from_code:
            logger.info(f"Broader BFS for base CUIs {cuis_from_code} ")

            defined_concepts = itertools.chain.from_iterable(
                definitions_search(cui) for cui in cuis_from_code
            )
            # since we searched multiple CUIs, ensure that they are returned in
            # order of closest to farthest. Also, need to ensure uniqueness.
            unique_defined_concepts = {c["ui"]: c for c in defined_concepts}

            if unique_defined_concepts:
                return sorted(
                    unique_defined_concepts.values(),
                    key=lambda _: _.get("distanceFromOrigin"),
                )
            logger.info("No broader concepts with definitions.")
        else:
            logger.info(f"UMLS concept not found for {source_vocab}/{concept_id}")

    # did not find the concept directly (by code)
    if source_desc:
        # if we have a source description, try to use it to find a CUI
        # Use strict matching if we were provided with a source code initially. This will happen if
        # the provided code is an MModal addition (not in original vocab).
        cleaned = text.remove_trailing_parens(source_desc)
        max_search_results = 5  # only check the top 5 results
        search_result = lookup_umls.term_search(
            api,
            term=cleaned,
            max_results=max_search_results,
            strict_match=bool(concept_id),
        )

        if search_result:
            search_term = search_result["searchTerm"]
            logger.info(f"Results for term search '{search_term}'")

            for concept in search_result["concepts"]:
                cui = concept["ui"]
                name = concept["name"]
                logger.info(f"Searching concept '{name}' {cui}")
                defs = definitions_search(cui)
                if defs:
                    return defs

    return []


def add_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Add arguments for search functions. Expected to be used with :meth:`find_builder`

    :param parser: the parser
    :return: same parser
    """
    df_group = parser.add_argument_group("Definitions Finder")

    # cui_search = df_group.add_argument_group("CUI Search")
    df_group.add_argument(
        "--start-cui", help="starting Concept ID", type=str, default=None
    )

    # sa_search = df_group.add_argument_group("Source Asserted Search")
    df_group.add_argument("--source-vocab", help="source vocab", type=str, default=None)
    df_group.add_argument(
        "--concept-id", help="source concept id", type=str, default=None
    )
    df_group.add_argument(
        "--source-desc", help="source concept description", type=str, default=None
    )

    df_group.add_argument(
        "--search-broader",
        help="search broader concepts. If false, search narrower.",
        type=str2bool,
        default=True,
        dest="broader",
    )
    df_group.add_argument(
        "--stop-on-found",
        help="stop searching after processing first level containing defined concepts",
        type=str2bool,
        default=True,
    )

    df_group.add_argument(
        "--max-distance",
        help="maximum allowed distance from `start_cui` (Default = 0 = Infinity)",
        type=int,
        default=0,
    )
    df_group.add_argument(
        "--language",
        help="only consider concepts in this language",
        type=str,
        default="ENG",
    )

    df_group.add_argument(
        "--preserve-semantic-type",
        help="preserve the semantic type assigned to start concept",
        type=str2bool,
        default=False,
    )

    return parser


_EXPECTED_KWARG_NAMES = vars(add_args(argparse.ArgumentParser()).parse_args([])).keys()


def find_builder(api: MetaThesaurus, parsed_args: argparse.Namespace):
    """
    Build a method to find defined concepts. Calls either :meth:`definitions_bfs` or
    :meth:`find_defined_concepts` depending on the args. Additional arguments are passed as keyword
    arguments. Such arguments override those passed with `parsed_args`

    :param api: MetaThesaurus API object
    :param parsed_args: parsed args. see :meth:`add_args`
    :return: function to find defined concepts
    """
    vargs = vars(parsed_args)
    # white list
    base_kwargs = {k: vargs[k] for k in _EXPECTED_KWARG_NAMES}
    # drop None
    base_kwargs = {k: v for k, v in base_kwargs.items() if v is not None}

    def find_fun(**kwargs: Any) -> List[Dict]:
        # override base kwargs
        merged_kwargs = base_kwargs.copy()
        merged_kwargs.update(kwargs)

        is_cui_search = "start_cui" in merged_kwargs
        is_source_search = (
            "source_vocab" in merged_kwargs
            or "concept_id" in merged_kwargs
            or "source_desc" in merged_kwargs
        )
        assert operator.xor(is_cui_search, is_source_search), (
            "Expected either 'start_cui' or some source asserted info such as "
            "'source_vocab', 'concept_id' 'source_desc' -- not both."
        )

        if is_cui_search:
            return definitions_bfs(api=api, **merged_kwargs)
        elif is_source_search:
            return find_defined_concepts(api=api, **merged_kwargs)
        else:
            raise AssertionError("Unreachable code")

    return find_fun


def definitions_itr(concepts: List[Dict]) -> Iterable[str]:
    """Iterate over definitions. Nearest to farthest. If there are multiple concepts at a given
    distance iterate over them, in a round-robin order"""
    # group by distance; anything closer will always come first
    for dist, group in itertools.groupby(
        concepts, key=lambda _: _["distanceFromOrigin"]
    ):
        defs = ((obj.get("value") for obj in _.get("definitions")) for _ in group)
        yield from iterators.roundrobin(*defs)



import argparse


# our copy
def str2bool(v):
    """ArgumentParser doesn't support type=bool. Thus, this helper method will convert
    from possible string types to True / False."""
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")



import logging
import os.path
from typing import Optional, Iterable, Callable, Any, List

logger = logging.getLogger(os.path.basename(__file__))


class FIFO(object):
    def __init__(
        self,
        iterable: Optional[Iterable] = None,
    ):
        if iterable:
            self._list = list(iterable)
        else:
            self._list = list()

    def push(self, item):
        self._list.append(item)

    def pop(self):
        item = self._list.pop(0)
        return item

    def peek(self):
        return self._list[0]

    def push_all(self, iterable):
        for item in iterable:
            self.push(item)

    def __iter__(self):
        return iter(self._list)

    def __len__(self):
        return len(self._list)

    def __contains__(self, item):
        return item in self._list

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self)


class UniqueFIFO(object):
    """
    FIFO queue that will not add elements which are already present.
    """

    def __init__(
        self,
        iterable: Optional[Iterable] = None,
        keyfn: Optional[Callable[[Any], Any]] = None,
    ):
        self._uniq = set()
        self._items = list()
        if keyfn:
            self._keyfn = keyfn
        else:
            self._keyfn = lambda _: _

        if iterable:
            self.push_all(iterable)

    @property
    def items(self) -> List:
        return self._items.copy()

    def push(self, item):
        key = self._keyfn(item)
        if key not in self._uniq:
            self._items.append(item)
            self._uniq.add(key)

    def pop(self):
        item = self._items.pop(0)
        self._uniq.discard(self._keyfn(item))
        return item

    def peek(self):
        return self._items[0]

    def push_all(self, iterable):
        for item in iterable:
            self.push(item)

    def remove(self, item):
        key = self._keyfn(item)
        if key not in self._uniq:
            raise KeyError(f"'{item}' not in set")
        self._items.remove(item)
        self._uniq.discard(key)

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)

    def __contains__(self, item):
        return self._keyfn(item) in self._uniq

    def __str__(self):
        return str(self._items)

    def __repr__(self):
        return str(self)



import re
from typing import List, Callable, Iterable

_ENCLOSING_CITATION = {")": "(", "]": "[", ">": "<"}


def remove_trailing_parens(text: str) -> str:
    # remove trailing parentheses e.g. "Room air (substance)" or "foo bar (klopfer: 2022)"
    close_char = text[-1]
    if close_char in _ENCLOSING_CITATION:
        stack = [close_char]
        open_char = _ENCLOSING_CITATION.get(close_char)
        i = len(text) - 2
        while i >= 0:
            cur_char = text[i]
            if cur_char == open_char:
                stack.pop()
                if not stack:
                    break
            elif cur_char == close_char:
                stack.append(close_char)

            i -= 1
        if i > 0:
            return text[:i]

    return text


def clean_definition_text(text: str) -> str:
    clean = text.strip()
    if not clean:
        return clean
    # clean trailing citation
    clean = remove_trailing_parens(clean)
    # xml tags
    clean = re.sub(r"<[^<]+?>", "", clean)

    return clean.strip()


_NOS_PAT = re.compile(r",?\s*NOS$")
_PUNCT_PAT = re.compile(r"\s*[,;:?_{|}.=()\[\]!/<>\"\\~-]\s*")


def normalize(text_str: str) -> str:
    """Remove common metadata strings, remove punctuation, lowercase, remove redundant whitespace"""
    # remove NOS (SNOMED ; Not Otherwise Specified)
    normalized = _NOS_PAT.sub("", text_str)
    normalized = clean_definition_text(normalized.lower())
    normalized = _PUNCT_PAT.sub(" ", normalized)
    normalized = re.sub(r"\s+", " ", normalized)
    return normalized.strip()


_STOP_WORDS = {"of", "the", "by"}


def tokenize(text_str: str) -> List[str]:
    """Tokenize string"""
    return [_ for _ in text_str.split() if _ and _ not in _STOP_WORDS]


def norm_tokenize(text_str: str) -> List[str]:
    """Normalize string, then tokenize"""
    return tokenize(normalize(text_str))


def _hammingish(source: Iterable[str], target: Iterable[str]) -> float:
    """This distance metric favors shorter 'target' sequences"""
    ss, st = set(source), set(target)

    dist = len(st - ss) / len(st)
    return dist


def hammingish(source: List[str], target: List[str]) -> float:
    return max(_hammingish(source, target), _hammingish(target, source))


def hammingish_str(source: str, target: str) -> float:
    return hammingish(norm_tokenize(source), norm_tokenize(target))


def hammingish_partial(source_txt: str) -> Callable[[str], float]:
    source = norm_tokenize(source_txt)

    def sort_key(target_txt: str) -> float:
        target = norm_tokenize(target_txt)
        return hammingish(source, target)

    return sort_key


def identity(text: str) -> str:
    return text



import argparse
import functools
import logging
import operator
from typing import Dict, Iterator, List, Optional, Tuple

from umlsrat import const
from umlsrat.api.rat_session import MetaThesaurusSession


class MetaThesaurus(object):
    """
    `UMLS MetaThesaurus API <https://documentation.uts.nlm.nih.gov/rest/home.html>`_ interface with caching.

    :param session: session object. If not provided, a default session will be constructed. See :py:meth:umlsrat.api.rat_session.MetaThesaurusSession.__init__
    :param umls_version: version of UMLS. By default, the specific latest version (e.g. 2023AA) is determined from the API.
    """

    def __init__(
        self,
        session: Optional[MetaThesaurusSession] = None,
        umls_version: Optional[str] = None,
    ):
        if session:
            self.session = session
        else:
            self.session = MetaThesaurusSession()

        if umls_version:
            self.umls_version = umls_version
        else:
            self.umls_version = self.session.get_current_version()
            self._logger.info("Using latest UMLS version '%s'", self.umls_version)

        self._rest_uri = const.UMLS_REST_ENDPOINT

        if self.session.use_cache and self.umls_version == "current":
            # may want to simply disable caching if version is 'current'
            self._logger.warning(
                "Version is 'current' and caching is enabled! "
                "API contents may change while the cache will not unless "
                "cleared."
            )

    @staticmethod
    def add_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Add constructor arguments to parser.

        :param parser: parser
        :return: the same parser
        """
        parser = MetaThesaurusSession.add_args(parser)

        group = parser.add_argument_group("MetaThesaurus")
        group.add_argument(
            "--umls-version",
            type=str,
            help="UMLS version to use. By default, fetch the exact current version string.",
            default=None,
        )

        return parser

    @staticmethod
    def from_namespace(args: argparse.Namespace) -> "MetaThesaurus":
        """
        Construct new object using values from namespace.

        :param args: parsed args
        :return: new instance
        """
        return MetaThesaurus(
            session=MetaThesaurusSession.from_namespace(args),
            umls_version=args.umls_version,
        )

    @property
    def _logger(self):
        return logging.getLogger(self.__class__.__name__)

    #### UMLS ####

    @property
    def _start_content_uri(self) -> str:
        """http://uts-ws.nlm.nih.gov/rest/content/{self.version}"""
        return f"{self._rest_uri}/content/{self.umls_version}"

    def get_concept(self, cui: str) -> Dict:
        """
        Get a UMLS concept by CUI.


        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> MetaThesaurus().get_concept("C0009044")

        .. code-block:: js

            {
            "ui": "C0009044",
            "name": "Closed fracture of carpal bone",
            "dateAdded": "09-30-1990",
            "majorRevisionDate": "03-16-2016",
            "classType": "Concept",
            "suppressible": false,
            "status": "R",
            "semanticTypes": [
             {
               "name": "Injury or Poisoning",
               "uri": "https://uts-ws.nlm.nih.gov/rest/semantic-network/2021AB/TUI/T037"
             }
            ],
            "atoms": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044/atoms",
            "definitions": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044/definitions",
            "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044/relations",
            "defaultPreferredAtom": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044/atoms/preferred",
            "atomCount": 71,
            "cvMemberCount": 0,
            "attributeCount": 0,
            "relationCount": 5
            }

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/concept/index.html>`__

        :param cui: Concept Unique Identifier (CUI) for the UMLS concept
        :return: Concept
        """
        assert cui
        uri = f"{self._start_content_uri}/CUI/{cui}"
        return self.session.get_single_result(uri)

    def get_definitions(
        self, cui: str, max_results: Optional[int] = None
    ) -> Iterator[Dict]:
        """
        Get the definitions for a concept.

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_definitions("C0009044"))

        .. code-block:: js

             [
               {
                 "rootSource": "NCI",
                 "value": "A traumatic break in one or more of the carpal bones that does not involve a break in the adjacent skin.",
                 "classType": "Definition",
                 "sourceOriginated": true
               }
             ]

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/definitions/index.html>`__

        :param cui: Concept Unique Identifier (CUI) for the UMLS concept
        :param max_results: maximum number of result to return. None = no max
        :return: generator yielding Definition info objects
        """
        uri = f"{self._start_content_uri}/CUI/{cui}/definitions"
        return self.session.get_results(uri, max_results=max_results)

    def get_relations(
        self, cui: str, max_results: Optional[int] = None, **params
    ) -> Iterator[Dict]:
        """
        Get relations for a concept

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/relations/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_relations("C0009044"))

        .. code-block:: js

            [
              {
                "ui": "R03033072",
                "suppressible": false,
                "sourceUi": null,
                "obsolete": false,
                "sourceOriginated": false,
                "rootSource": "MTH",
                "relationLabel": "RO",
                "additionalRelationLabel": "",
                "groupId": null,
                "relatedId": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009353",
                "relatedIdName": "Colles' Fracture",
                "classType": "ConceptRelation",
                "attributeCount": 0
              },
              ...
            ]

        :param cui: Concept Unique Identifier (CUI) for the UMLS concept
        :param max_results: maximum number of result to return. None = no max
        :return: generator yielding Relation info objects
        """
        uri = f"{self._start_content_uri}/CUI/{cui}/relations"
        return self.session.get_results(uri, max_results=max_results, **params)

    def get_atoms_for_cui(
        self, cui: str, max_results: Optional[int] = None, **params
    ) -> Iterator[Dict]:
        """
        Get Atoms for a concept.

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/atoms/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_atoms_for_cui(cui="C0009044", language="ENG"))

        .. code-block:: js

            [
              {
                "classType": "Atom",
                "ui": "A0243916",
                "sourceDescriptor": null,
                "sourceConcept": null,
                "concept": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044",
                "obsolete": "false",
                "suppressible": "false",
                "rootSource": "RCD",
                "termType": "PT",
                "code": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/RCD/S240.",
                "language": "ENG",
                "name": "Closed fracture of carpal bone",
                "attributes": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/attributes",
                "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/relations",
                "children": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/children",
                "descendants": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/descendants",
                "parents": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/parents",
                "ancestors": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/ancestors",
                "contentViewMemberships": [
                  {
                    "memberUri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357/member/A0243916",
                    "name": "MetaMap NLP View",
                    "uri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357"
                  }
                ]
              },
              ...
            ]

        :param cui: Concept Unique Identifier (CUI) for the UMLS concept
        :param max_results: maximum number of result to return. None = no max
        :param params: additional params passed to the GET request See `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/atoms/index.html>`__
        :return: list of Relation info objects
        """
        uri = f"{self._start_content_uri}/CUI/{cui}/atoms"
        return self.session.get_results(uri, max_results=max_results, **params)

    def get_atom(self, aui: str) -> Optional[Dict]:
        """
        Get Atoms for a concept.

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/atoms/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_atom(aui="A0243916"))

        .. code-block:: js

            {
              "classType": "Atom",
              "ui": "A0243916",
              "sourceDescriptor": null,
              "sourceConcept": null,
              "concept": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0009044",
              "obsolete": "false",
              "suppressible": "false",
              "rootSource": "RCD",
              "termType": "PT",
              "code": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/RCD/S240.",
              "language": "ENG",
              "name": "Closed fracture of carpal bone",
              "attributes": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/attributes",
              "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/relations",
              "children": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/children",
              "descendants": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/descendants",
              "parents": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/parents",
              "ancestors": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0243916/ancestors",
              "contentViewMemberships": [
                {
                  "memberUri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357/member/A0243916",
                  "name": "MetaMap NLP View",
                  "uri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357"
                }
              ]
            }

        :param aui: Atom Unique Identifier (AUI)
        :return: atom information
        """
        url = f"{self._start_content_uri}/AUI/{aui}"
        return self.session.get_single_result(url)

    def get_atom_ancestors(
        self, aui: str, max_results: Optional[int] = None
    ) -> Iterator[Dict]:
        """
        Get ancestors of an Atom

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/atoms/ancestors-and-descendants/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_atom_ancestors(aui="A0243916"))

        .. code-block:: js

            [
              {
                "classType": "Atom",
                "ui": "A0004658",
                "sourceDescriptor": null,
                "sourceConcept": null,
                "concept": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0016658",
                "obsolete": "false",
                "suppressible": "false",
                "rootSource": "RCD",
                "termType": "PT",
                "code": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/RCD/XA0FK",
                "language": "ENG",
                "name": "Fracture",
                "attributes": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/attributes",
                "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/relations",
                "children": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/children",
                "descendants": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/descendants",
                "parents": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/parents",
                "ancestors": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/AUI/A0004658/ancestors",
                "contentViewMemberships": [
                  {
                    "memberUri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357/member/A0004658",
                    "name": "MetaMap NLP View",
                    "uri": "https://uts-ws.nlm.nih.gov/rest/content-views/2021AB/CUI/C1700357"
                  }
                ]
              },
              ...
            ]

        :param aui: Atom Unique Identifier (AUI) for the UMLS Atom
        :param max_results: maximum number of result to return. None = no max
        :return: generator yielding Concepts
        """
        assert aui.startswith("A"), f"Invalid AUI '{aui}'"
        uri = f"{self._start_content_uri}/AUI/{aui}/ancestors"
        return self.session.get_results(uri, max_results=max_results)

    ###
    # Search
    ###

    def search(
        self, query: str, max_results: Optional[int] = None, **params
    ) -> Iterator[Dict]:
        """
        Search UMLS!

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/search/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> MetaThesaurus().search("cheese", max_results=1)

        .. code-block:: js

            [
              {
                "ui": "C0007968",
                "rootSource": "MTH",
                "uri": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/CUI/C0007968",
                "name": "Cheese"
              }
            ]

        :param query: search string
        :param max_results: maximum number of result to return. None = no max
        :param params: additional get request params
        :return: generator yielding search results
        """
        uri = f"https://uts-ws.nlm.nih.gov/rest/search/{self.umls_version}"
        if "string" in params:
            self._logger.warning(
                "Overwriting existing 'string' value %s", params["string"]
            )
        params["string"] = query
        return self.session.get_results(uri, max_results=max_results, **params)

    ###
    # Source Asserted
    ####

    def get_source_concept(self, source_vocab: str, concept_id: str) -> Optional[Dict]:
        """
        Get a source asserted concept. e.g. a concept from SNOMED, LOINC, etc

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> MetaThesaurus().get_source_concept(source_vocab="snomed", concept_id="75508005")

        .. code-block:: js

            {
              "classType": "SourceAtomCluster",
              "ui": "75508005",
              "suppressible": false,
              "obsolete": false,
              "rootSource": "SNOMEDCT_US",
              "atomCount": 2,
              "cVMemberCount": 0,
              "attributes": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/attributes",
              "atoms": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/atoms",
              "ancestors": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/ancestors",
              "parents": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/parents",
              "children": null,
              "descendants": null,
              "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/relations",
              "definitions": null,
              "concepts": "https://uts-ws.nlm.nih.gov/rest/search/2021AB?string=75508005&sabs=SNOMEDCT_US&searchType=exact&inputType=sourceUi",
              "defaultPreferredAtom": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/SNOMEDCT_US/75508005/atoms/preferred",
              "subsetMemberships": [
                {
                  "memberUri": "https://uts-ws.nlm.nih.gov/rest/subsets/2021AB/source/SNOMEDCT_US/900000000000497000/member/75508005",
                  "uri": "https://uts-ws.nlm.nih.gov/rest/subsets/2021AB/source/SNOMEDCT_US/900000000000497000",
                  "name": "CTV3 simple map"
                }
              ],
              "contentViewMemberships": [],
              "name": "Dissecting"
            }

        :param source_vocab: source Vocabulary
        :param concept_id: concept ID
        :return: concept or None
        """
        source_vocab = self.validate_source_abbrev(source_vocab)
        assert concept_id
        uri = f"{self._start_content_uri}/source/{source_vocab}/{concept_id}"
        return self.session.get_single_result(uri)

    def get_source_relations(
        self,
        source_vocab: str,
        concept_id: str,
        **params,
    ) -> Iterator[Dict]:
        """
        Get source asserted relations

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/relations/index.html>`__

        .. code-block:: python

            from umlsrat.api.metathesaurus import MetaThesaurus
            list(MetaThesaurus().get_source_relations(source_vocab="MSH",
                                                      concept_id="D002415",
                                                      includeRelationLabels="RN,CHD",))


        .. code-block:: js

            [
              {
                "ui": "R71237095",
                "suppressible": false,
                "sourceUi": null,
                "obsolete": false,
                "sourceOriginated": false,
                "rootSource": "MSH",
                "groupId": null,
                "attributeCount": 0,
                "classType": "AtomClusterRelation",
                "relatedFromId": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002415",
                "relatedFromIdName": "Cats",
                "relationLabel": "CHD",
                "additionalRelationLabel": "",
                "relatedId": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991",
                "relatedIdName": "Felis"
              }
            ]

        :param source_vocab: source Vocabulary
        :param concept_id: concept ID
        :return: generator yielding concepts
        """
        source_vocab = self.validate_source_abbrev(source_vocab)
        assert concept_id
        uri = f"{self._start_content_uri}/source/{source_vocab}/{concept_id}/relations"

        return self.session.get_results(uri, **params)

    def get_source_parents(
        self,
        source_vocab: str,
        concept_id: str,
    ) -> Iterator[Dict]:
        """
        Get source asserted parents

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/parents-and-children/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_source_parents(source_vocab="MSH", concept_id="D002415"))


        .. code-block:: python

            [
              {
                "classType": "SourceAtomCluster",
                "ui": "D045991",
                "suppressible": false,
                "obsolete": false,
                "rootSource": "MSH",
                "atomCount": 1,
                "cVMemberCount": 0,
                "attributes": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/attributes",
                "atoms": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/atoms",
                "ancestors": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/ancestors",
                "parents": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/parents",
                "children": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/children",
                "descendants": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/descendants",
                "relations": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/relations",
                "definitions": null,
                "concepts": "https://uts-ws.nlm.nih.gov/rest/search/2021AB?string=D045991&sabs=MSH&searchType=exact&inputType=sourceUi",
                "defaultPreferredAtom": "https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D045991/atoms/preferred",
                "subsetMemberships": [],
                "contentViewMemberships": [],
                "name": "Felis"
              }
            ]

        :param source_vocab: source Vocabulary
        :param concept_id: concept ID
        :return: generator yielding concepts
        """
        source_vocab = self.validate_source_abbrev(source_vocab)
        assert concept_id
        uri = f"{self._start_content_uri}/source/{source_vocab}/{concept_id}/parents"

        return self.session.get_results(uri)

    def get_source_ancestors(
        self,
        source_vocab: str,
        concept_id: str,
    ) -> Iterator[Dict]:
        """
        Get source asserted ancestors

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/ancestors-and-descendants/index.html>`__

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> list(MetaThesaurus().get_source_ancestors(source_vocab="MSH", concept_id="D002415"))

        .. code-block:: python

            [{'ancestors': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/ancestors',
              'atomCount': 1,
              'atoms': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/atoms',
              'attributes': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/attributes',
              'cVMemberCount': 0,
              'children': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/children',
              'classType': 'SourceAtomCluster',
              'concepts': 'https://uts-ws.nlm.nih.gov/rest/search/2021AB?string=D002335&sabs=MSH&searchType=exact&inputType=sourceUi',
              'contentViewMemberships': [],
              'defaultPreferredAtom': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/atoms/preferred',
              'definitions': None,
              'descendants': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/descendants',
              'name': 'Carnivora',
              'obsolete': False,
              'parents': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/parents',
              'relations': 'https://uts-ws.nlm.nih.gov/rest/content/2021AB/source/MSH/D002335/relations',
              'rootSource': 'MSH',
              'subsetMemberships': [],
              'suppressible': False,
              'ui': 'D002335'},
             ...]

        :param source_vocab: source Vocabulary
        :param concept_id: concept ID
        :return: generator yielding concepts
        """
        source_vocab = self.validate_source_abbrev(source_vocab)
        assert concept_id
        uri = f"{self._start_content_uri}/source/{source_vocab}/{concept_id}/ancestors"

        return self.session.get_results(uri)

    def crosswalk(
        self, source_vocab: str, concept_id: str, language: str = None, **params
    ) -> Iterator[Dict]:
        """
        Crosswalk vocabularies. Find all source atoms which share the same CUI.

        `UMLS Doc <https://documentation.uts.nlm.nih.gov/rest/source-asserted-identifiers/crosswalk/>`__


        :param source_vocab: source Vocabulary
        :param concept_id: concept ID
        :param language: target language
        :return: generator yielding atom clusters
        """
        source_vocab = self.validate_source_abbrev(source_vocab)
        assert concept_id
        uri = f"{self._rest_uri}/crosswalk/{self.umls_version}/source/{source_vocab}/{concept_id}"

        if language:
            params["targetSource"] = self.get_sabs_str(
                language, params.pop("targetSource", None)
            )

        return self.session.get_results(uri, **params)

    ##
    # Source vocabulary metadata and lookup.
    ##
    @property
    def _source_metadata(self) -> Iterator[Dict]:
        """
        Get metadata for UMLS sources.

        :return: list of metadata about sources
        """
        url = f"{self._rest_uri}/metadata/{self.umls_version}/sources"
        return self.session.get_results(url)

    @property
    @functools.lru_cache()
    def source_metadata_index(self) -> Dict[str, Dict]:
        """
        Get source metadata indexed by abbreviation.

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> MetaThesaurus().source_metadata_index.get("LNC")

        .. code-block: js

            {
              "classType": "RootSource",
              "abbreviation": "LNC",
              "expandedForm": "LOINC, 271",
              "family": "LNC",
              "language": {
                "classType": "Language",
                "abbreviation": "ENG",
                "expandedForm": "English"
              },
              "restrictionLevel": 0,
              ...
            }

        :return: source metadata indexed by abbreviation
        """
        index = {}
        for datum in self._source_metadata:
            abbr = datum["abbreviation"]
            assert abbr not in index
            index[abbr] = datum
        return index

    def find_source_info(self, name: str, fuzzy: bool = False) -> Optional[Dict]:
        """
        Find metadata for a source vocabulary.

        :param name: the abbreviation or name of the vocab
        :param fuzzy: do a fuzzy match on non abbreviation fields (i.e. "shortName")
        :return: source info or None if it cannot be found
        """
        data = self.source_metadata_index.get(name)
        if data:
            return data

        abbr_upper = name.upper()
        data = self.source_metadata_index.get(abbr_upper)
        if data:
            return data

        abbr_mapped = const._COMMON_VOCAB_NAMES.get(abbr_upper)
        data = self.source_metadata_index.get(abbr_mapped)
        if data:
            return data

        if not fuzzy:
            return None

        # fuzzy matching
        candidates = list(self.source_metadata_index.values())

        def match_field(field_name: str, op):
            results = list()
            for c in candidates:
                if c[field_name] and op(c[field_name].upper(), abbr_upper):
                    results.append(c)
            if not results:
                return None

            if len(results) == 1:
                return results.pop()
            else:
                raise ValueError(
                    f"Ambiguous vocab abbr '{name}' -- fuzzy match returned {len(results)} results.\n"
                    f"{[_[field_name] for _ in results]}"
                )

        data = match_field("shortName", operator.eq)
        if data:
            return data

        data = match_field("shortName", operator.contains)
        if data:
            return data

        # no match
        return None

    def validate_source_abbrev(self, sab: str) -> str:
        """
        Look up the source info for the abbreviation. If it exists, return the properly normalized
        abbreviation. Otherwise, raise.

        .. code-block: python
            from umlsrat.api.metathesaurus import MetaThesaurus

            MetaThesaurus().validate_source_abbrev("snomed")
            # "SNOMEDCT_US"

            MetaThesaurus().validate_source_abbrev("loinx")
            # raises ValueError

        :param sab: source abbreviation
        :return: normalized source abbreviation
        :raises ValueError: when abbreviation is invalid
        """
        info = self.find_source_info(sab)
        if not info:
            raise ValueError(f"No such vocabulary abbreviation '{sab}'")
        else:
            return info["abbreviation"]

    @functools.lru_cache(maxsize=2)
    def sources_for_language(self, lab: str) -> Tuple[str]:
        """
        Get a list of source abbreviations associated with a given language (abbreviation).

        >>> from umlsrat.api.metathesaurus import MetaThesaurus
        >>> MetaThesaurus().sources_for_language("GER")

        .. code-block: js

            ('DMDICD10', 'DMDUMD', 'ICPCGER', 'LNC-DE-AT',
                'LNC-DE-DE', 'MDRGER', 'MSHGER', 'WHOGER')

        :param lab: language abbreviation
        :return: list of source abbreviations
        """
        sabs_list = [
            source["abbreviation"]
            for source in self.source_metadata_index.values()
            if source["language"]["abbreviation"] == lab
        ]  # type: List[str]
        return tuple(sorted(sabs_list))

    def validate_language_abbrev(self, lab: str) -> str:
        """
        Look up the sources for language. If any exist, return the properly normalized language
        abbreviation. Otherwise, raise.

        .. code-block: python
            from umlsrat.api.metathesaurus import MetaThesaurus

            MetaThesaurus().validate_language_abbrev("fre")
            # "FRE"

            MetaThesaurus().validate_language_abbrev("FRA")
            # raises ValueError

        :param lab: language abbreviation
        :return: normalized abbreviation
        :raises ValueError: if no sources exist for the language
        """
        assert lab, "must supply language abbreviation"
        abbr_upper = lab.upper()
        if self.sources_for_language(abbr_upper):
            return abbr_upper
        raise ValueError(f"No sources found for language abbrev '{lab}'")

    def _valid_ordered_split_sabs(self, sabs_str: str) -> List[str]:
        # maintaining order is important for caching
        return sorted(
            set(self.validate_source_abbrev(_.strip()) for _ in sabs_str.split(","))
        )

    def get_sabs_str(self, language: str, sabs: Optional[str] = None) -> str:
        """
        Return a comma-separated list of source abbreviations for a target language. If asabs string
        is provided, we simply validate that it belongs to the target language.

        :param language: target language
        :param sabs: comma-separated list of source abbreviations
        :return:
        """
        if not language:
            if not sabs:
                return ""
            else:
                # simply validate the sabs
                return ",".join(self._valid_ordered_split_sabs(sabs))

        lab = self.validate_language_abbrev(language)
        lang_sab_tpl = self.sources_for_language(lab)

        if sabs:
            existing = self._valid_ordered_split_sabs(sabs)
            existing_set = set(existing)
            lang_sab_set = set(lang_sab_tpl)
            not_in_language = existing_set - lang_sab_set
            if not_in_language:
                raise ValueError(
                    f"{not_in_language} source vocabularies are not associated with language '{language}'"
                )
            # so if you have specified sabs already, we just validate them and send them back
            return sabs
        else:
            return ",".join(lang_sab_tpl)
